@@ -20,9 +20,13 @@
 */
 
 #include "incidence.h"
+#include "incidence_p.h"
 #include "calformat.h"
+#include "kcalendarcore_debug.h"
 #include "utils_p.h"
 
+#include <math.h>
+
 #include <QStringList>
 #include <QTextDocument> // for .toHtmlEscaped() and Qt::mightBeRichText()
 #include <QTime>
@@ -30,154 +34,131 @@
 
 using namespace KCalendarCore;
 
-/**
-  Private class that helps to provide binary compatibility between releases.
-  @internal
-*/
-//@cond PRIVATE
-class Q_DECL_HIDDEN KCalendarCore::Incidence::Private
+IncidencePrivate::IncidencePrivate() = default;
+
+IncidencePrivate::IncidencePrivate(const IncidencePrivate &p)
+    : IncidenceBasePrivate(p)
+    , mCreated(p.mCreated)
+    , mDescription(p.mDescription)
+    , mSummary(p.mSummary)
+    , mLocation(p.mLocation)
+    , mCategories(p.mCategories)
+    , mResources(p.mResources)
+    , mStatusString(p.mStatusString)
+    , mSchedulingID(p.mSchedulingID)
+    , mRelatedToUid(p.mRelatedToUid)
+    , mRecurrenceId(p.mRecurrenceId)
+    , mConferences(p.mConferences)
+    , mGeoLatitude(p.mGeoLatitude)
+    , mGeoLongitude(p.mGeoLongitude)
+    , mRecurrence(nullptr)
+    , mRevision(p.mRevision)
+    , mPriority(p.mPriority)
+    , mStatus(p.mStatus)
+    , mSecrecy(p.mSecrecy)
+    , mColor(p.mColor)
+    , mDescriptionIsRich(p.mDescriptionIsRich)
+    , mSummaryIsRich(p.mSummaryIsRich)
+    , mLocationIsRich(p.mLocationIsRich)
+    , mThisAndFuture(p.mThisAndFuture)
+    , mLocalOnly(false)
 {
-public:
-    Private()
-        : mGeoLatitude(INVALID_LATLON)
-        , mGeoLongitude(INVALID_LATLON)
-        , mRecurrence(nullptr)
-        , mRevision(0)
-        , mPriority(0)
-        , mStatus(StatusNone)
-        , mSecrecy(SecrecyPublic)
-        , mDescriptionIsRich(false)
-        , mSummaryIsRich(false)
-        , mLocationIsRich(false)
-        , mHasGeo(false)
-        , mThisAndFuture(false)
-        , mLocalOnly(false)
-    {
+}
+
+IncidencePrivate::IncidencePrivate(const Incidence &other)
+    : IncidencePrivate(*other.d_func())
+{
+}
+
+void IncidencePrivate::clear()
+{
+    mAlarms.clear();
+    mAttachments.clear();
+    delete mRecurrence;
+    mRecurrence = nullptr;
+}
+
+void IncidencePrivate::init(Incidence *q, const IncidencePrivate &other)
+{
+    mRevision = other.mRevision;
+    mCreated = other.mCreated;
+    mDescription = other.mDescription;
+    mDescriptionIsRich = other.mDescriptionIsRich;
+    mSummary = other.mSummary;
+    mSummaryIsRich = other.mSummaryIsRich;
+    mCategories = other.mCategories;
+    mRelatedToUid = other.mRelatedToUid;
+    mResources = other.mResources;
+    mStatusString = other.mStatusString;
+    mStatus = other.mStatus;
+    mSecrecy = other.mSecrecy;
+    mPriority = other.mPriority;
+    mLocation = other.mLocation;
+    mLocationIsRich = other.mLocationIsRich;
+    mGeoLatitude = other.mGeoLatitude;
+    mGeoLongitude = other.mGeoLongitude;
+    mRecurrenceId = other.mRecurrenceId;
+    mConferences = other.mConferences;
+    mThisAndFuture = other.mThisAndFuture;
+    mLocalOnly = other.mLocalOnly;
+    mColor = other.mColor;
+
+    // Alarms and Attachments are stored in ListBase<...>, which is a QValueList<...*>.
+    // We need to really duplicate the objects stored therein, otherwise deleting
+    // i will also delete all attachments from this object (setAutoDelete...)
+    mAlarms.reserve(other.mAlarms.count());
+    for (const Alarm::Ptr &alarm : qAsConst(other.mAlarms)) {
+        Alarm::Ptr b(new Alarm(*alarm.data()));
+        b->setParent(q);
+        mAlarms.append(b);
     }
 
-    Private(const Private &p)
-        : mCreated(p.mCreated)
-        , mDescription(p.mDescription)
-        , mSummary(p.mSummary)
-        , mLocation(p.mLocation)
-        , mCategories(p.mCategories)
-        , mResources(p.mResources)
-        , mStatusString(p.mStatusString)
-        , mSchedulingID(p.mSchedulingID)
-        , mRelatedToUid(p.mRelatedToUid)
-        , mRecurrenceId(p.mRecurrenceId)
-        , mConferences(p.mConferences)
-        , mGeoLatitude(p.mGeoLatitude)
-        , mGeoLongitude(p.mGeoLongitude)
-        , mRecurrence(nullptr)
-        , mRevision(p.mRevision)
-        , mPriority(p.mPriority)
-        , mStatus(p.mStatus)
-        , mSecrecy(p.mSecrecy)
-        , mColor(p.mColor)
-        , mDescriptionIsRich(p.mDescriptionIsRich)
-        , mSummaryIsRich(p.mSummaryIsRich)
-        , mLocationIsRich(p.mLocationIsRich)
-        , mHasGeo(p.mHasGeo)
-        , mThisAndFuture(p.mThisAndFuture)
-        , mLocalOnly(false)
-    {
-    }
-
-    void clear()
-    {
-        mAlarms.clear();
-        mAttachments.clear();
-        delete mRecurrence;
+    mAttachments = other.mAttachments;
+    if (other.mRecurrence) {
+        mRecurrence = new Recurrence(*(other.mRecurrence));
+        mRecurrence->addObserver(q);
+    } else {
         mRecurrence = nullptr;
     }
+}
 
-    void init(Incidence *dest, const Incidence &src)
-    {
-        mRevision = src.d->mRevision;
-        mCreated = src.d->mCreated;
-        mDescription = src.d->mDescription;
-        mSummary = src.d->mSummary;
-        mCategories = src.d->mCategories;
-        mRelatedToUid = src.d->mRelatedToUid;
-        mResources = src.d->mResources;
-        mStatusString = src.d->mStatusString;
-        mStatus = src.d->mStatus;
-        mSecrecy = src.d->mSecrecy;
-        mPriority = src.d->mPriority;
-        mLocation = src.d->mLocation;
-        mGeoLatitude = src.d->mGeoLatitude;
-        mGeoLongitude = src.d->mGeoLongitude;
-        mHasGeo = src.d->mHasGeo;
-        mRecurrenceId = src.d->mRecurrenceId;
-        mConferences = src.d->mConferences;
-        mThisAndFuture = src.d->mThisAndFuture;
-        mLocalOnly = src.d->mLocalOnly;
-
-        // Alarms and Attachments are stored in ListBase<...>, which is a QValueList<...*>.
-        // We need to really duplicate the objects stored therein, otherwise deleting
-        // i will also delete all attachments from this object (setAutoDelete...)
-        mAlarms.reserve(src.d->mAlarms.count());
-        for (const Alarm::Ptr &alarm : qAsConst(src.d->mAlarms)) {
-            Alarm::Ptr b(new Alarm(*alarm.data()));
-            b->setParent(dest);
-            mAlarms.append(b);
-        }
-
-        mAttachments = src.d->mAttachments;
-        if (src.d->mRecurrence) {
-            mRecurrence = new Recurrence(*(src.d->mRecurrence));
-            mRecurrence->addObserver(dest);
-        } else {
-            mRecurrence = nullptr;
-        }
-    }
-
-    QDateTime mCreated; // creation datetime
-    QString mDescription; // description string
-    QString mSummary; // summary string
-    QString mLocation; // location string
-    QStringList mCategories; // category list
-    Attachment::List mAttachments; // attachments list
-    Alarm::List mAlarms; // alarms list
-    QStringList mResources; // resources list (not calendar resources)
-    QString mStatusString; // status string, for custom status
-    QString mSchedulingID; // ID for scheduling mails
-    QMap<RelType, QString> mRelatedToUid; // incidence uid this is related to, for each relType
-    QDateTime mRecurrenceId; // recurrenceId
-    Conference::List mConferences; // conference list
-
-    float mGeoLatitude; // Specifies latitude in decimal degrees
-    float mGeoLongitude; // Specifies longitude in decimal degrees
-    mutable Recurrence *mRecurrence; // recurrence
-    int mRevision; // revision number
-    int mPriority; // priority: 1 = highest, 2 = less, etc.
-    Status mStatus; // status
-    Secrecy mSecrecy; // secrecy
-    QString mColor; // background color
-    bool mDescriptionIsRich = false; // description string is richtext.
-    bool mSummaryIsRich = false; // summary string is richtext.
-    bool mLocationIsRich = false; // location string is richtext.
-    bool mHasGeo = false; // if incidence has geo data
-    bool mThisAndFuture = false;
-    bool mLocalOnly = false; // allow changes that won't go to the server
-};
+bool IncidencePrivate::validStatus(Incidence::Status status)
+{
+    return status == Incidence::StatusNone;
+}
 //@endcond
 
+#if KCALENDARCORE_BUILD_DEPRECATED_SINCE(5, 91)
 Incidence::Incidence()
-    : IncidenceBase()
-    , d(new KCalendarCore::Incidence::Private)
+    : IncidenceBase(new IncidencePrivate)
 {
     recreate();
     resetDirtyFields();
 }
 
 Incidence::Incidence(const Incidence &i)
-    : IncidenceBase(i)
+    : IncidenceBase(i, new IncidencePrivate(*(i.d_func())))
     , Recurrence::RecurrenceObserver()
-    , d(new KCalendarCore::Incidence::Private(*i.d))
 {
-    d->init(this, i);
+    Q_D(Incidence);
+    d->init(this, *(i.d_func()));
+    resetDirtyFields();
+}
+#endif
+
+Incidence::Incidence(IncidencePrivate *p)
+    : IncidenceBase(p)
+{
+    recreate();
+    resetDirtyFields();
+}
+
+Incidence::Incidence(const Incidence &other, IncidencePrivate *p)
+    : IncidenceBase(other, p)
+    , Recurrence::RecurrenceObserver()
+{
+    Q_D(Incidence);
+    d->init(this, *(other.d_func()));
     resetDirtyFields();
 }
 
@@ -185,11 +166,11 @@ Incidence::~Incidence()
 {
     // Alarm has a raw incidence pointer, so we must set it to 0
     // so Alarm doesn't use it after Incidence is destroyed
+    Q_D(const Incidence);
     for (const Alarm::Ptr &alarm : qAsConst(d->mAlarms)) {
         alarm->setParent(nullptr);
     }
     delete d->mRecurrence;
-    delete d;
 }
 
 //@cond PRIVATE
@@ -202,12 +183,13 @@ static bool stringCompare(const QString &s1, const QString &s2)
 //@endcond
 IncidenceBase &Incidence::assign(const IncidenceBase &other)
 {
+    Q_D(Incidence);
     if (&other != this) {
         d->clear();
         // TODO: should relations be cleared out, as in destructor???
         IncidenceBase::assign(other);
         const Incidence *i = static_cast<const Incidence *>(&other);
-        d->init(this, *i);
+        d->init(this, *(i->d_func()));
     }
 
     return *this;
@@ -222,54 +204,45 @@ bool Incidence::equals(const IncidenceBase &incidence) const
     // If they weren't the same type IncidenceBase::equals would had returned false already
     const Incidence *i2 = static_cast<const Incidence *>(&incidence);
 
-    const auto alarmList = alarms();
-    const auto otherAlarmsList = i2->alarms();
+    const Alarm::List alarmList = alarms();
+    const Alarm::List otherAlarmsList = i2->alarms();
     if (alarmList.count() != otherAlarmsList.count()) {
         return false;
     }
 
-    Alarm::List::ConstIterator a1 = alarmList.constBegin();
-    Alarm::List::ConstIterator a1end = alarmList.constEnd();
-    Alarm::List::ConstIterator a2 = otherAlarmsList.constBegin();
-    Alarm::List::ConstIterator a2end = otherAlarmsList.constEnd();
-    for (; a1 != a1end && a2 != a2end; ++a1, ++a2) {
-        if (**a1 == **a2) {
-            continue;
-        } else {
-            return false;
-        }
+    auto matchFunc = [](const Alarm::Ptr &a, const Alarm::Ptr &b) {
+        return *a == *b;
+    };
+
+    const auto [it1, it2] = std::mismatch(alarmList.cbegin(), alarmList.cend(), otherAlarmsList.cbegin(), otherAlarmsList.cend(), matchFunc);
+    // Checking the iterator from one list only, since both lists are the same size
+    if (it1 != alarmList.cend()) {
+        return false;
     }
 
-    const auto attachmentList = attachments();
-    const auto otherAttachmentList = i2->attachments();
+    const Attachment::List attachmentList = attachments();
+    const Attachment::List otherAttachmentList = i2->attachments();
     if (attachmentList.count() != otherAttachmentList.count()) {
         return false;
     }
 
-    Attachment::List::ConstIterator att1 = attachmentList.constBegin();
-    const Attachment::List::ConstIterator att1end = attachmentList.constEnd();
-    Attachment::List::ConstIterator att2 = otherAttachmentList.constBegin();
-    const Attachment::List::ConstIterator att2end = otherAttachmentList.constEnd();
-    for (; att1 != att1end && att2 != att2end; ++att1, ++att2) {
-        if (*att1 == *att2) {
-            continue;
-        } else {
-            return false;
-        }
+    const auto [at1, at2] = std::mismatch(attachmentList.cbegin(), attachmentList.cend(), otherAttachmentList.cbegin(), otherAttachmentList.cend());
+
+    // Checking the iterator from one list only, since both lists are the same size
+    if (at1 != attachmentList.cend()) {
+        return false;
     }
 
-    bool recurrenceEqual = (d->mRecurrence == nullptr && i2->d->mRecurrence == nullptr);
+    Q_D(const Incidence);
+    bool recurrenceEqual = (d->mRecurrence == nullptr && i2->d_func()->mRecurrence == nullptr);
     if (!recurrenceEqual) {
         recurrence(); // create if doesn't exist
         i2->recurrence(); // create if doesn't exist
-        recurrenceEqual = d->mRecurrence != nullptr && i2->d->mRecurrence != nullptr && *d->mRecurrence == *i2->d->mRecurrence;
+        recurrenceEqual = d->mRecurrence != nullptr && i2->d_func()->mRecurrence != nullptr
+            && *d->mRecurrence == *i2->d_func()->mRecurrence;
     }
 
-    if (d->mHasGeo == i2->d->mHasGeo) {
-        if (d->mHasGeo && (!qFuzzyCompare(d->mGeoLatitude, i2->d->mGeoLatitude) || !qFuzzyCompare(d->mGeoLongitude, i2->d->mGeoLongitude))) {
-            return false;
-        }
-    } else {
+    if (!qFuzzyCompare(d->mGeoLatitude, i2->d_func()->mGeoLatitude) || !qFuzzyCompare(d->mGeoLongitude, i2->d_func()->mGeoLongitude)) {
         return false;
     }
     // clang-format off
@@ -277,15 +250,18 @@ bool Incidence::equals(const IncidenceBase &incidence) const
         recurrenceEqual
         && created() == i2->created()
         && stringCompare(description(), i2->description())
+        && descriptionIsRich() == i2->descriptionIsRich()
         && stringCompare(summary(), i2->summary())
+        && summaryIsRich() == i2->summaryIsRich()
         && categories() == i2->categories()
         && stringCompare(relatedTo(), i2->relatedTo())
         && resources() == i2->resources()
-        && d->mStatus == i2->d->mStatus
-        && (d->mStatus == StatusNone || stringCompare(d->mStatusString, i2->d->mStatusString))
+        && d->mStatus == i2->d_func()->mStatus
+        && (d->mStatus == StatusNone || stringCompare(d->mStatusString, i2->d_func()->mStatusString))
         && secrecy() == i2->secrecy()
         && priority() == i2->priority()
         && stringCompare(location(), i2->location())
+        && locationIsRich() == i2->locationIsRich()
         && stringCompare(color(), i2->color())
         && stringCompare(schedulingID(), i2->schedulingID())
         && recurrenceId() == i2->recurrenceId()
@@ -314,6 +290,7 @@ void Incidence::recreate()
 
 void Incidence::setLastModified(const QDateTime &lm)
 {
+    Q_D(const Incidence);
     if (!d->mLocalOnly) {
         IncidenceBase::setLastModified(lm);
     }
@@ -321,6 +298,7 @@ void Incidence::setLastModified(const QDateTime &lm)
 
 void Incidence::setReadOnly(bool readOnly)
 {
+    Q_D(const Incidence);
     IncidenceBase::setReadOnly(readOnly);
     if (d->mRecurrence) {
         d->mRecurrence->setRecurReadOnly(readOnly);
@@ -332,11 +310,13 @@ void Incidence::setLocalOnly(bool localOnly)
     if (mReadOnly) {
         return;
     }
+    Q_D(Incidence);
     d->mLocalOnly = localOnly;
 }
 
 bool Incidence::localOnly() const
 {
+    Q_D(const Incidence);
     return d->mLocalOnly;
 }
 
@@ -345,6 +325,7 @@ void Incidence::setAllDay(bool allDay)
     if (mReadOnly) {
         return;
     }
+    Q_D(const Incidence);
     if (d->mRecurrence) {
         d->mRecurrence->setAllDay(allDay);
     }
@@ -353,33 +334,34 @@ void Incidence::setAllDay(bool allDay)
 
 void Incidence::setCreated(const QDateTime &created)
 {
-    if (mReadOnly || d->mLocalOnly) {
-        return;
-    }
-
-    d->mCreated = created.toUTC();
-    const auto ct = d->mCreated.time();
-    // Remove milliseconds
-    d->mCreated.setTime(QTime(ct.hour(), ct.minute(), ct.second()));
-    setFieldDirty(FieldCreated);
-
-    // FIXME: Shouldn't we call updated for the creation date, too?
-    //  updated();
-}
-
-QDateTime Incidence::created() const
-{
-    return d->mCreated;
-}
-
-void Incidence::setRevision(int rev)
-{
+    Q_D(Incidence);
     if (mReadOnly || d->mLocalOnly) {
         return;
     }
 
     update();
+    d->mCreated = created.toUTC();
+    const auto ct = d->mCreated.time();
+    // Remove milliseconds
+    d->mCreated.setTime(QTime(ct.hour(), ct.minute(), ct.second()));
+    setFieldDirty(FieldCreated);
+    updated();
+}
 
+QDateTime Incidence::created() const
+{
+    Q_D(const Incidence);
+    return d->mCreated;
+}
+
+void Incidence::setRevision(int rev)
+{
+    Q_D(Incidence);
+    if (mReadOnly || d->mLocalOnly) {
+        return;
+    }
+
+    update();
     d->mRevision = rev;
     setFieldDirty(FieldRevision);
     updated();
@@ -387,11 +369,13 @@ void Incidence::setRevision(int rev)
 
 int Incidence::revision() const
 {
+    Q_D(const Incidence);
     return d->mRevision;
 }
 
 void Incidence::setDtStart(const QDateTime &dt)
 {
+    Q_D(const Incidence);
     IncidenceBase::setDtStart(dt);
     if (d->mRecurrence && dirtyFields().contains(FieldDtStart)) {
         d->mRecurrence->setStartDateTime(dt, allDay());
@@ -400,12 +384,18 @@ void Incidence::setDtStart(const QDateTime &dt)
 
 void Incidence::shiftTimes(const QTimeZone &oldZone, const QTimeZone &newZone)
 {
+    Q_D(const Incidence);
     IncidenceBase::shiftTimes(oldZone, newZone);
     if (d->mRecurrence) {
         d->mRecurrence->shiftTimes(oldZone, newZone);
     }
-    for (int i = 0, end = d->mAlarms.count(); i < end; ++i) {
-        d->mAlarms[i]->shiftTimes(oldZone, newZone);
+    if (d->mAlarms.count() > 0) {
+        update();
+        for (auto alarm : d->mAlarms) {
+            alarm->shiftTimes(oldZone, newZone);
+        }
+        setFieldDirty(FieldAlarms);
+        updated();
     }
 }
 
@@ -415,6 +405,7 @@ void Incidence::setDescription(const QString &description, bool isRich)
         return;
     }
     update();
+    Q_D(Incidence);
     d->mDescription = description;
     d->mDescriptionIsRich = isRich;
     setFieldDirty(FieldDescription);
@@ -428,11 +419,13 @@ void Incidence::setDescription(const QString &description)
 
 QString Incidence::description() const
 {
+    Q_D(const Incidence);
     return d->mDescription;
 }
 
 QString Incidence::richDescription() const
 {
+    Q_D(const Incidence);
     if (descriptionIsRich()) {
         return d->mDescription;
     } else {
@@ -442,6 +435,7 @@ QString Incidence::richDescription() const
 
 bool Incidence::descriptionIsRich() const
 {
+    Q_D(const Incidence);
     return d->mDescriptionIsRich;
 }
 
@@ -450,6 +444,7 @@ void Incidence::setSummary(const QString &summary, bool isRich)
     if (mReadOnly) {
         return;
     }
+    Q_D(Incidence);
     if (d->mSummary != summary || d->mSummaryIsRich != isRich) {
         update();
         d->mSummary = summary;
@@ -466,11 +461,13 @@ void Incidence::setSummary(const QString &summary)
 
 QString Incidence::summary() const
 {
+    Q_D(const Incidence);
     return d->mSummary;
 }
 
 QString Incidence::richSummary() const
 {
+    Q_D(const Incidence);
     if (summaryIsRich()) {
         return d->mSummary;
     } else {
@@ -480,6 +477,7 @@ QString Incidence::richSummary() const
 
 bool Incidence::summaryIsRich() const
 {
+    Q_D(const Incidence);
     return d->mSummaryIsRich;
 }
 
@@ -489,8 +487,10 @@ void Incidence::setCategories(const QStringList &categories)
         return;
     }
 
+    Q_D(Incidence);
     update();
     d->mCategories = categories;
+    setFieldDirty(FieldCategories);
     updated();
 }
 
@@ -502,6 +502,7 @@ void Incidence::setCategories(const QString &catStr)
     update();
     setFieldDirty(FieldCategories);
 
+    Q_D(Incidence);
     d->mCategories.clear();
 
     if (catStr.isEmpty()) {
@@ -511,9 +512,8 @@ void Incidence::setCategories(const QString &catStr)
 
     d->mCategories = catStr.split(QLatin1Char(','));
 
-    QStringList::Iterator it;
-    for (it = d->mCategories.begin(); it != d->mCategories.end(); ++it) {
-        *it = (*it).trimmed();
+    for (auto &category : d->mCategories) {
+        category = category.trimmed();
     }
 
     updated();
@@ -521,11 +521,13 @@ void Incidence::setCategories(const QString &catStr)
 
 QStringList Incidence::categories() const
 {
+    Q_D(const Incidence);
     return d->mCategories;
 }
 
 QString Incidence::categoriesStr() const
 {
+    Q_D(const Incidence);
     return d->mCategories.join(QLatin1Char(','));
 }
 
@@ -534,6 +536,7 @@ void Incidence::setRelatedTo(const QString &relatedToUid, RelType relType)
     // TODO: RFC says that an incidence can have more than one related-to field
     // even for the same relType.
 
+    Q_D(Incidence);
     if (d->mRelatedToUid[relType] != relatedToUid) {
         update();
         d->mRelatedToUid[relType] = relatedToUid;
@@ -544,6 +547,7 @@ void Incidence::setRelatedTo(const QString &relatedToUid, RelType relType)
 
 QString Incidence::relatedTo(RelType relType) const
 {
+    Q_D(const Incidence);
     return d->mRelatedToUid.value(relType);
 }
 
@@ -552,6 +556,7 @@ void Incidence::setColor(const QString &colorName)
     if (mReadOnly) {
         return;
     }
+    Q_D(Incidence);
     if (!stringCompare(d->mColor, colorName)) {
         update();
         d->mColor = colorName;
@@ -562,6 +567,7 @@ void Incidence::setColor(const QString &colorName)
 
 QString Incidence::color() const
 {
+    Q_D(const Incidence);
     return d->mColor;
 }
 
@@ -569,6 +575,7 @@ QString Incidence::color() const
 
 Recurrence *Incidence::recurrence() const
 {
+    Q_D(const Incidence);
     if (!d->mRecurrence) {
         d->mRecurrence = new Recurrence();
         d->mRecurrence->setStartDateTime(dateTime(RoleRecurrenceStart), allDay());
@@ -582,12 +589,14 @@ Recurrence *Incidence::recurrence() const
 
 void Incidence::clearRecurrence()
 {
+    Q_D(const Incidence);
     delete d->mRecurrence;
     d->mRecurrence = nullptr;
 }
 
 ushort Incidence::recurrenceType() const
 {
+    Q_D(const Incidence);
     if (d->mRecurrence) {
         return d->mRecurrence->recurrenceType();
     } else {
@@ -597,6 +606,7 @@ ushort Incidence::recurrenceType() const
 
 bool Incidence::recurs() const
 {
+    Q_D(const Incidence);
     if (d->mRecurrence) {
         return d->mRecurrence->recurs();
     } else {
@@ -606,11 +616,13 @@ bool Incidence::recurs() const
 
 bool Incidence::recursOn(const QDate &date, const QTimeZone &timeZone) const
 {
+    Q_D(const Incidence);
     return d->mRecurrence && d->mRecurrence->recursOn(date, timeZone);
 }
 
 bool Incidence::recursAt(const QDateTime &qdt) const
 {
+    Q_D(const Incidence);
     return d->mRecurrence && d->mRecurrence->recursAt(qdt);
 }
 
@@ -629,7 +641,7 @@ QList<QDateTime> Incidence::startDateTimesForDate(const QDate &date, const QTime
     // if the incidence doesn't recur,
     QDateTime kdate(date, {}, timeZone);
     if (!recurs()) {
-        if (!(start > kdate || end < kdate)) {
+        if (start.date() <= date && end.date() >= date) {
             result << start;
         }
         return result;
@@ -714,6 +726,7 @@ void Incidence::addAttachment(const Attachment &attachment)
         return;
     }
 
+    Q_D(Incidence);
     update();
     d->mAttachments.append(attachment);
     setFieldDirty(FieldAttachment);
@@ -722,25 +735,27 @@ void Incidence::addAttachment(const Attachment &attachment)
 
 void Incidence::deleteAttachments(const QString &mime)
 {
-    Attachment::List result;
-    Attachment::List::Iterator it = d->mAttachments.begin();
-    while (it != d->mAttachments.end()) {
-        if ((*it).mimeType() != mime) {
-            result += *it;
-        }
-        ++it;
+    Q_D(Incidence);
+    auto it = std::remove_if(d->mAttachments.begin(), d->mAttachments.end(), [&mime](const Attachment &a) {
+        return a.mimeType() == mime;
+    });
+    if (it != d->mAttachments.end()) {
+        update();
+        d->mAttachments.erase(it, d->mAttachments.end());
+        setFieldDirty(FieldAttachment);
+        updated();
     }
-    d->mAttachments = result;
-    setFieldDirty(FieldAttachment);
 }
 
 Attachment::List Incidence::attachments() const
 {
+    Q_D(const Incidence);
     return d->mAttachments;
 }
 
 Attachment::List Incidence::attachments(const QString &mime) const
 {
+    Q_D(const Incidence);
     Attachment::List attachments;
     for (const Attachment &attachment : qAsConst(d->mAttachments)) {
         if (attachment.mimeType() == mime) {
@@ -752,8 +767,11 @@ Attachment::List Incidence::attachments(const QString &mime) const
 
 void Incidence::clearAttachments()
 {
+    Q_D(Incidence);
+    update();
     setFieldDirty(FieldAttachment);
     d->mAttachments.clear();
+    updated();
 }
 
 void Incidence::setResources(const QStringList &resources)
@@ -763,6 +781,7 @@ void Incidence::setResources(const QStringList &resources)
     }
 
     update();
+    Q_D(Incidence);
     d->mResources = resources;
     setFieldDirty(FieldResources);
     updated();
@@ -770,6 +789,7 @@ void Incidence::setResources(const QStringList &resources)
 
 QStringList Incidence::resources() const
 {
+    Q_D(const Incidence);
     return d->mResources;
 }
 
@@ -779,7 +799,13 @@ void Incidence::setPriority(int priority)
         return;
     }
 
+    if (priority < 0 || priority > 9) {
+        qCWarning(KCALCORE_LOG)  << "Ignoring invalid priority" << priority;
+        return;
+    }
+
     update();
+    Q_D(Incidence);
     d->mPriority = priority;
     setFieldDirty(FieldPriority);
     updated();
@@ -787,20 +813,27 @@ void Incidence::setPriority(int priority)
 
 int Incidence::priority() const
 {
+    Q_D(const Incidence);
     return d->mPriority;
 }
 
 void Incidence::setStatus(Incidence::Status status)
 {
-    if (mReadOnly || status == StatusX) {
+    if (mReadOnly) {
+        qCWarning(KCALCORE_LOG)  << "Attempt to set status of read-only incidence";
         return;
     }
 
-    update();
-    d->mStatus = status;
-    d->mStatusString.clear();
-    setFieldDirty(FieldStatus);
-    updated();
+    Q_D(Incidence);
+    if (d->validStatus(status)) {
+        update();
+        d->mStatus = status;
+        d->mStatusString.clear();
+        setFieldDirty(FieldStatus);
+        updated();
+    } else {
+        qCWarning(KCALCORE_LOG)  << "Ignoring invalid status" << status << "for" << typeStr();
+    }
 }
 
 void Incidence::setCustomStatus(const QString &status)
@@ -810,6 +843,7 @@ void Incidence::setCustomStatus(const QString &status)
     }
 
     update();
+    Q_D(Incidence);
     d->mStatus = status.isEmpty() ? StatusNone : StatusX;
     d->mStatusString = status;
     setFieldDirty(FieldStatus);
@@ -818,11 +852,13 @@ void Incidence::setCustomStatus(const QString &status)
 
 Incidence::Status Incidence::status() const
 {
+    Q_D(const Incidence);
     return d->mStatus;
 }
 
 QString Incidence::customStatus() const
 {
+    Q_D(const Incidence);
     if (d->mStatus == StatusX) {
         return d->mStatusString;
     } else {
@@ -837,6 +873,7 @@ void Incidence::setSecrecy(Incidence::Secrecy secrecy)
     }
 
     update();
+    Q_D(Incidence);
     d->mSecrecy = secrecy;
     setFieldDirty(FieldSecrecy);
     updated();
@@ -844,23 +881,27 @@ void Incidence::setSecrecy(Incidence::Secrecy secrecy)
 
 Incidence::Secrecy Incidence::secrecy() const
 {
+    Q_D(const Incidence);
     return d->mSecrecy;
 }
 
 Alarm::List Incidence::alarms() const
 {
+    Q_D(const Incidence);
     return d->mAlarms;
 }
 
 Alarm::Ptr Incidence::newAlarm()
 {
+    Q_D(Incidence);
     Alarm::Ptr alarm(new Alarm(this));
-    d->mAlarms.append(alarm);
+    addAlarm(alarm);
     return alarm;
 }
 
 void Incidence::addAlarm(const Alarm::Ptr &alarm)
 {
+    Q_D(Incidence);
     update();
     d->mAlarms.append(alarm);
     setFieldDirty(FieldAlarms);
@@ -869,6 +910,7 @@ void Incidence::addAlarm(const Alarm::Ptr &alarm)
 
 void Incidence::removeAlarm(const Alarm::Ptr &alarm)
 {
+    Q_D(Incidence);
     const int index = d->mAlarms.indexOf(alarm);
     if (index > -1) {
         update();
@@ -881,6 +923,7 @@ void Incidence::removeAlarm(const Alarm::Ptr &alarm)
 void Incidence::clearAlarms()
 {
     update();
+    Q_D(Incidence);
     d->mAlarms.clear();
     setFieldDirty(FieldAlarms);
     updated();
@@ -888,22 +931,22 @@ void Incidence::clearAlarms()
 
 bool Incidence::hasEnabledAlarms() const
 {
-    for (const Alarm::Ptr &alarm : qAsConst(d->mAlarms)) {
-        if (alarm->enabled()) {
-            return true;
-        }
-    }
-    return false;
+    Q_D(const Incidence);
+    return std::any_of(d->mAlarms.cbegin(), d->mAlarms.cend(), [](const Alarm::Ptr &alarm) {
+        return alarm->enabled();
+    });
 }
 
 Conference::List Incidence::conferences() const
 {
+    Q_D(const Incidence);
     return d->mConferences;
 }
 
 void Incidence::addConference(const Conference &conference)
 {
     update();
+    Q_D(Incidence);
     d->mConferences.push_back(conference);
     setFieldDirty(FieldConferences);
     updated();
@@ -912,6 +955,7 @@ void Incidence::addConference(const Conference &conference)
 void Incidence::setConferences(const Conference::List &conferences)
 {
     update();
+    Q_D(Incidence);
     d->mConferences = conferences;
     setFieldDirty(FieldConferences);
     updated();
@@ -920,6 +964,7 @@ void Incidence::setConferences(const Conference::List &conferences)
 void Incidence::clearConferences()
 {
     update();
+    Q_D(Incidence);
     d->mConferences.clear();
     setFieldDirty(FieldConferences);
     updated();
@@ -931,6 +976,7 @@ void Incidence::setLocation(const QString &location, bool isRich)
         return;
     }
 
+    Q_D(Incidence);
     if (d->mLocation != location || d->mLocationIsRich != isRich) {
         update();
         d->mLocation = location;
@@ -947,11 +993,13 @@ void Incidence::setLocation(const QString &location)
 
 QString Incidence::location() const
 {
+    Q_D(const Incidence);
     return d->mLocation;
 }
 
 QString Incidence::richLocation() const
 {
+    Q_D(const Incidence);
     if (locationIsRich()) {
         return d->mLocation;
     } else {
@@ -961,6 +1009,7 @@ QString Incidence::richLocation() const
 
 bool Incidence::locationIsRich() const
 {
+    Q_D(const Incidence);
     return d->mLocationIsRich;
 }
 
@@ -969,14 +1018,18 @@ void Incidence::setSchedulingID(const QString &sid, const QString &uid)
     if (!uid.isEmpty()) {
         setUid(uid);
     }
+    Q_D(Incidence);
     if (sid != d->mSchedulingID) {
+        update();
         d->mSchedulingID = sid;
         setFieldDirty(FieldSchedulingId);
+        updated();
     }
 }
 
 QString Incidence::schedulingID() const
 {
+    Q_D(const Incidence);
     if (d->mSchedulingID.isNull()) {
         // Nothing set, so use the normal uid
         return uid();
@@ -986,29 +1039,34 @@ QString Incidence::schedulingID() const
 
 bool Incidence::hasGeo() const
 {
-    return d->mHasGeo;
+    Q_D(const Incidence);
+    // For internal consistency, return false if either coordinate is invalid.
+    return d->mGeoLatitude != INVALID_LATLON && d->mGeoLongitude != INVALID_LATLON;
 }
-
+#if KCALENDARCORE_BUILD_DEPRECATED_SINCE(5, 89)
 void Incidence::setHasGeo(bool hasGeo)
 {
     if (mReadOnly) {
         return;
     }
 
-    if (hasGeo == d->mHasGeo) {
-        return;
+    Q_D(Incidence);
+    if (!hasGeo) {
+        update();
+        d->mGeoLatitude = INVALID_LATLON;
+        d->mGeoLongitude = INVALID_LATLON;
+        setFieldDirty(FieldGeoLatitude);
+        setFieldDirty(FieldGeoLongitude);
+        updated();
     }
-
-    update();
-    d->mHasGeo = hasGeo;
-    setFieldDirty(FieldGeoLatitude);
-    setFieldDirty(FieldGeoLongitude);
-    updated();
+    // If hasGeo is true, the caller should set latitude and longitude to legal values..
 }
-
+#endif
 float Incidence::geoLatitude() const
 {
-    return d->mGeoLatitude;
+    Q_D(const Incidence);
+    // For internal consistency, both coordinates are considered invalid if either is.
+    return (INVALID_LATLON == d->mGeoLongitude) ? INVALID_LATLON : d->mGeoLatitude;
 }
 
 void Incidence::setGeoLatitude(float geolatitude)
@@ -1017,6 +1075,15 @@ void Incidence::setGeoLatitude(float geolatitude)
         return;
     }
 
+    if (isnan(geolatitude)) {
+        geolatitude = INVALID_LATLON;
+    }
+    if (geolatitude != INVALID_LATLON && (geolatitude < -90.0 || geolatitude > 90.0)) {
+        qCWarning(KCALCORE_LOG) << "Ignoring invalid  latitude" << geolatitude;
+        return;
+    }
+
+    Q_D(Incidence);
     update();
     d->mGeoLatitude = geolatitude;
     setFieldDirty(FieldGeoLatitude);
@@ -1025,36 +1092,53 @@ void Incidence::setGeoLatitude(float geolatitude)
 
 float Incidence::geoLongitude() const
 {
-    return d->mGeoLongitude;
+    Q_D(const Incidence);
+    // For internal consistency, both coordinates are considered invalid if either is.
+    return (INVALID_LATLON == d->mGeoLatitude) ? INVALID_LATLON : d->mGeoLongitude;
 }
 
 void Incidence::setGeoLongitude(float geolongitude)
 {
-    if (!mReadOnly) {
-        update();
-        d->mGeoLongitude = geolongitude;
-        setFieldDirty(FieldGeoLongitude);
-        updated();
+    if (mReadOnly) {
+        return;
     }
+
+    if (isnan(geolongitude)) {
+        geolongitude = INVALID_LATLON;
+    }
+    if (geolongitude != INVALID_LATLON && (geolongitude < -180.0 || geolongitude > 180.0)) {
+        qCWarning(KCALCORE_LOG) << "Ignoring invalid  longitude" << geolongitude;
+        return;
+    }
+
+    Q_D(Incidence);
+    update();
+    d->mGeoLongitude = geolongitude;
+    setFieldDirty(FieldGeoLongitude);
+    updated();
 }
 
 bool Incidence::hasRecurrenceId() const
 {
+    Q_D(const Incidence);
     return (allDay() && d->mRecurrenceId.date().isValid()) || d->mRecurrenceId.isValid();
 }
 
 QDateTime Incidence::recurrenceId() const
 {
+    Q_D(const Incidence);
     return d->mRecurrenceId;
 }
 
 void Incidence::setThisAndFuture(bool thisAndFuture)
 {
+    Q_D(Incidence);
     d->mThisAndFuture = thisAndFuture;
 }
 
 bool Incidence::thisAndFuture() const
 {
+    Q_D(const Incidence);
     return d->mThisAndFuture;
 }
 
@@ -1062,6 +1146,7 @@ void Incidence::setRecurrenceId(const QDateTime &recurrenceId)
 {
     if (!mReadOnly) {
         update();
+        Q_D(Incidence);
         d->mRecurrenceId = recurrenceId;
         setFieldDirty(FieldRecurrenceId);
         updated();
@@ -1073,6 +1158,7 @@ void Incidence::setRecurrenceId(const QDateTime &recurrenceId)
     belongs to. */
 void Incidence::recurrenceUpdated(Recurrence *recurrence)
 {
+    Q_D(const Incidence);
     if (recurrence == d->mRecurrence) {
         update();
         setFieldDirty(FieldRecurrence);
@@ -1120,12 +1206,14 @@ QStringList Incidence::mimeTypes()
 
 void Incidence::serialize(QDataStream &out) const
 {
+    Q_D(const Incidence);
     serializeQDateTimeAsKDateTime(out, d->mCreated);
     out << d->mRevision << d->mDescription << d->mDescriptionIsRich << d->mSummary << d->mSummaryIsRich << d->mLocation << d->mLocationIsRich << d->mCategories
-        << d->mResources << d->mStatusString << d->mPriority << d->mSchedulingID << d->mGeoLatitude << d->mGeoLongitude << d->mHasGeo;
+        << d->mResources << d->mStatusString << d->mPriority << d->mSchedulingID << d->mGeoLatitude << d->mGeoLongitude
+        << hasGeo();    // No longer used, but serialized/deserialized for compatibility.
     serializeQDateTimeAsKDateTime(out, d->mRecurrenceId);
-    out << d->mThisAndFuture << d->mLocalOnly << d->mStatus << d->mSecrecy << (d->mRecurrence ? true : false) << d->mAttachments.count() << d->mAlarms.count()
-        << d->mConferences.count() << d->mRelatedToUid;
+    out << d->mThisAndFuture << d->mLocalOnly << d->mStatus << d->mSecrecy << (d->mRecurrence ? true : false) << (qint32)d->mAttachments.count()
+        << (qint32)d->mAlarms.count() << (qint32)d->mConferences.count() << d->mRelatedToUid;
 
     if (d->mRecurrence) {
         out << d->mRecurrence;
@@ -1146,13 +1234,18 @@ void Incidence::serialize(QDataStream &out) const
 
 void Incidence::deserialize(QDataStream &in)
 {
-    quint32 status, secrecy;
+    bool hasGeo;    // No longer used, but serialized/deserialized for compatibility.
+    quint32 status;
+    quint32 secrecy;
     bool hasRecurrence;
-    int attachmentCount, alarmCount, conferencesCount;
+    int attachmentCount;
+    int alarmCount;
+    int conferencesCount;
     QMap<int, QString> relatedToUid;
+    Q_D(Incidence);
     deserializeKDateTimeAsQDateTime(in, d->mCreated);
     in >> d->mRevision >> d->mDescription >> d->mDescriptionIsRich >> d->mSummary >> d->mSummaryIsRich >> d->mLocation >> d->mLocationIsRich >> d->mCategories
-        >> d->mResources >> d->mStatusString >> d->mPriority >> d->mSchedulingID >> d->mGeoLatitude >> d->mGeoLongitude >> d->mHasGeo;
+        >> d->mResources >> d->mStatusString >> d->mPriority >> d->mSchedulingID >> d->mGeoLatitude >> d->mGeoLongitude >> hasGeo;
     deserializeKDateTimeAsQDateTime(in, d->mRecurrenceId);
     in >> d->mThisAndFuture >> d->mLocalOnly >> status >> secrecy >> hasRecurrence >> attachmentCount >> alarmCount >> conferencesCount >> relatedToUid;
 
@@ -1192,7 +1285,8 @@ void Incidence::deserialize(QDataStream &in)
 
     d->mRelatedToUid.clear();
 
-    auto it = relatedToUid.cbegin(), end = relatedToUid.cend();
+    auto it = relatedToUid.cbegin();
+    auto end = relatedToUid.cend();
     for (; it != end; ++it) {
         d->mRelatedToUid.insert(static_cast<Incidence::RelType>(it.key()), it.value());
     }
@@ -1215,10 +1309,12 @@ QVariantList toVariantList(int size, typename QVector<T>::ConstIterator begin, t
 
 QVariantList Incidence::attachmentsVariant() const
 {
+    Q_D(const Incidence);
     return toVariantList<Attachment>(d->mAttachments.size(), d->mAttachments.cbegin(), d->mAttachments.cend());
 }
 
 QVariantList Incidence::conferencesVariant() const
 {
+    Q_D(const Incidence);
     return toVariantList<Conference>(d->mConferences.size(), d->mConferences.cbegin(), d->mConferences.cend());
 }
